@@ -1,4 +1,6 @@
 #include <hildon/hildon.h>
+#include <pthread.h>
+#include <poll.h>
 
 #include "../common.h"
 
@@ -47,10 +49,12 @@ static int combo_keystate = 0;
 static int accel_keystate = 0;
 static int touch_keystate = 0;
 static int touch_idle = TRUE;
+static pthread_mutex_t keybd_mutex;
 
 // Map hardware key codes to local key numbers
-static unsigned char keymap[65536];
+static unsigned char keymap[512];
 static int use_zoom_keys = FALSE;
+static int use_gpio_keys = FALSE;
 
 // Accelerometer parameters
 static int accel_enable = FALSE;
@@ -214,6 +218,7 @@ focus_handler(GtkWindow *window)
   if (gtk_window_is_active(window)) {
     // Focus received, resume the game
     gtk_dialog_response(pause_dialog, 0);
+    use_gpio_keys = TRUE;
   } else {
     // Focus lost, but there could be different reasons to it:
     // + screen is locked (pause the game)
@@ -259,9 +264,17 @@ pause_keybd_handler(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
     gtk_dialog_response(pause_dialog, 0);
 }
 
+static void
+activate_tasknav(GtkWidget *widget, gpointer user_data)
+{
+    system("dbus-send /com/nokia/hildon_desktop com.nokia.hildon_desktop.exit_app_view");
+}
+
 static void activate_pause(int automatic)
 {
-  if (pause_dialog) return;
+  use_gpio_keys = FALSE;
+
+  if (pause_dialog || (automatic && !autopause)) return;
 
   if (automatic) {
     pause_dialog = GTK_DIALOG (hildon_note_new_information (GTK_WINDOW (window), "Paused"));
@@ -274,14 +287,22 @@ static void activate_pause(int automatic)
                                                             NULL));
 
     // Create mute button
-    GtkWidget *mute_button = hildon_check_button_new (HILDON_SIZE_AUTO_WIDTH | HILDON_SIZE_FINGER_HEIGHT);
+    GtkWidget *mute_button = hildon_check_button_new (HILDON_SIZE_FINGER_HEIGHT);
     gtk_button_set_label (GTK_BUTTON (mute_button), "Mute");
     hildon_check_button_set_active (HILDON_CHECK_BUTTON (mute_button), !global_enable_audio);
 
-    // Add mute button
+    // Create tasknav button
+    GtkWidget *tasknav_button = hildon_button_new (HILDON_SIZE_FINGER_HEIGHT, HILDON_BUTTON_ARRANGEMENT_VERTICAL);
+    gtk_button_set_label (GTK_BUTTON (tasknav_button), "Tasks");
+    g_signal_connect (tasknav_button, "clicked", G_CALLBACK (activate_tasknav), NULL);
+
+    // Populate the dialog
+    GtkWidget *hbox = gtk_hbox_new (TRUE, 0);
     GtkWidget *content_area = gtk_dialog_get_content_area (GTK_DIALOG (pause_dialog));
-    gtk_container_add (GTK_CONTAINER (content_area), mute_button);
-    gtk_widget_show (mute_button);
+    gtk_container_add (GTK_CONTAINER (hbox), mute_button);
+    gtk_container_add (GTK_CONTAINER (hbox), tasknav_button);
+    gtk_container_add (GTK_CONTAINER (content_area), hbox);
+    gtk_widget_show_all(hbox);
 
     g_signal_connect (G_OBJECT (pause_dialog), "key-release-event", G_CALLBACK (pause_keybd_handler), NULL);
 
@@ -321,10 +342,14 @@ keybd_handler(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
       return;
   }
 
+  pthread_mutex_lock(&keybd_mutex);
+
   if (event->type == GDK_KEY_PRESS)
     *target_keystate |= key;
   else if (event->type == GDK_KEY_RELEASE)
     *target_keystate &= ~key;
+
+  pthread_mutex_unlock(&keybd_mutex);
 }
 
 static void
@@ -388,7 +413,7 @@ delete_handler()
   quit();
 }
 
-void print_help()
+static void print_help()
 {
   printf("gpSP - Game Boy Advance emulator"                                         "\n"
          "Usage: gpspm FILE [options]"                                              "\n"
@@ -409,17 +434,59 @@ void print_help()
   exit(1);
 }
 
-void read_keys(FILE *keys_file)
+static void *gpio_watcher(void *arg)
+{
+  struct pollfd fd;
+  fd.fd = open("/sys/devices/platform/gpio-switch/cam_focus/state", O_RDONLY);
+  fd.events = POLLPRI;
+
+  // Currently only one key is available
+  u32 key = 1 << keymap[300+68];
+
+  while (TRUE) {
+    poll(&fd, 1, -1);
+
+    char state;
+    lseek(fd.fd, 0, SEEK_SET);
+    read(fd.fd, &state, 1);
+
+    if (!use_gpio_keys) continue;
+
+    pthread_mutex_lock(&keybd_mutex);
+
+    // 'a' is for "active"
+    if (state == 'a')
+      keybd_keystate |= key;
+    else
+      keybd_keystate &= ~key;
+
+    pthread_mutex_unlock(&keybd_mutex);
+  }
+
+  return NULL;
+}
+
+static void read_keys(FILE *keys_file)
 {
   // Bind keys like described in the provided configuration file
-  int i, key_code;
+  int i;
   for (i = 0; i < 20; i++) {
+    int key_code;
     fscanf(keys_file, "%i", &key_code);
     keymap[key_code] = i;
 
     // Check whether zoom keys should be captured
     if (key_code == 73 || key_code == 74)
         use_zoom_keys = TRUE;
+
+    // Check whether GPIO switches should be watched
+    if (key_code == 300+68)
+        use_gpio_keys = TRUE;
+  }
+
+  if (use_gpio_keys) {
+    pthread_t gpio_thread;
+    pthread_create(&gpio_thread, NULL, gpio_watcher, NULL);
   }
 }
 
@@ -545,6 +612,8 @@ void gpsp_plat_init(void)
   regions_ver[REG_A]          = gdk_region_rectangle(&rect_ver_a);
   regions_ver[REG_B]          = gdk_region_rectangle(&rect_ver_b);
 
+  pthread_mutex_init(&keybd_mutex, NULL);
+
   // Build the configuration directory path and export as the main path
   char *config_home = getenv("XDG_CONFIG_HOME");
   if (config_home == NULL || config_home[0] == 0)
@@ -554,7 +623,7 @@ void gpsp_plat_init(void)
 
   // Assign the default (do-nothing) value to all keys
   int i;
-  for (i = 0; i < 65536; i++)
+  for (i = 0; i < 512; i++)
       keymap[i] = 0xff;
 
   char filename[512];
@@ -614,7 +683,7 @@ void gpsp_plat_init(void)
   g_signal_connect (G_OBJECT (window), "motion-notify-event", G_CALLBACK (touch_motion_handler), NULL);
   g_signal_connect (G_OBJECT (window), "configure-event", G_CALLBACK (configure_handler), NULL);
   g_signal_connect (G_OBJECT (window), "delete-event", G_CALLBACK (delete_handler), NULL);
-  if (autopause) {
+  if (autopause || use_gpio_keys) {
     g_signal_connect (G_OBJECT (window), "notify::is-active", G_CALLBACK (focus_handler), NULL);
     g_signal_connect (G_OBJECT (window), "notify::is-topmost", G_CALLBACK (topmost_handler), NULL);
   }
